#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <cerrno>
#include <cstdlib>
#include <sstream>
#include <signal.h>
#include <unistd.h>

#include <ardour/ardour.h>

#include <ardour/tempo.h>
#include <ardour/audio_track.h>
#include <ardour/session_playlist.h>
#include <ardour/diskstream.h>
#include <ardour/playlist.h>
#include <ardour/audioplaylist.h>
#include <ardour/audioregion.h>
#include <ardour/region.h>
#include <ardour/region_factory.h>
#include <ardour/audiofilesource.h>
#include <ardour/session.h>
#include <ardour/connection.h>
#include <ardour/session_connection.h>

#include <dae/dae_msg_types.h>
#include <dae/dae_song_props.h>

#define DAE_NOT_LIB
#define CPLUSPLUS

extern int debug;
extern int zero;
namespace DAE {
	template<class T, class P> class MixerTrack;
}
#include "app.h"
#include "shared.h"
#include "dae/utils.h"
#include "dae/msgs.h"
#include "dae/msgs2.h"

namespace DAE {
#include <dae/dae_time.h>
#include <dae/dae_ardour.h>
using namespace DAE;

extern daeMessaging* core;
extern struct _shm_song* shm_index;
}
extern "C" {
	#include "dae/dae_shm.h"
}
namespace DAE {

using namespace ARDOUR;

void
dae_msg_set_obj(unsigned _trans_id, int* frag, short int* off, unsigned char* msg)
{
	/*
	a message has been received, requesting that a property be set for an object.

	note: idx is currently only 32bit, so we cannot use ardour 64bit id's.

	strictly speaking, all time positions should be in mu, though i'm not sure they all are atm.

	*/
	printf("%s()...\n", __func__);
	unsigned char* _data = msg;
	msg_printf(msg);

	int err=0, prop;

	//struct _song_pos val_sp;
	unsigned long long val_u64;
	unsigned val_ui;
	int val_i;
	short int val_si;
	unsigned short int val_usi=0;
	unsigned char val_uc=0;

	int block_num = 0;
	int m;             // secondary block index

	int return_idx = 0;

	debug=1;
	// 1 = stime 8 byte struct
	// 2 = len 8 byte len
	// 3-5 = pb_??? 4 byte signed
	// 6-45 = rest = 4 byte unsigned

	//if (debug) printf("songmsg_set_obj(%u,%d,%d,<data>);\n", _trans_id, *frag, *off);

	// u4 : idx
#ifdef NON_POD
	xmlNode* idx = NULL;
#else
	unsigned idx = 0;
#endif
	//printf("songmsg_set_obj(): msg=%p\n", msg);
	memcpy(&idx, &msg[0], 4);
	if (debug) printf("%s(): idx=%x\n", __func__, idx);

	uint32_t object_type = msg_get_obj_type(idx);
	printf("%s(): object_type=%x\n", __func__, object_type);
	idx = idx & ~ARDOUR_TYPE_MASK;
	//char test = msg[0];
	//printf("test=%u\n", test);

	//convert the ardour mask to dae type, see dae_song_client.h (FIXME ardour mask should probably be derived from dae type)
	unsigned char _type;
	if     (object_type==ARDOUR_TYPE_TRACK) _type = DAE_OBJECT_TRACK;
	else if(object_type==ARDOUR_TYPE_SONG ) _type = DAE_OBJECT_SONG;
	else{
		_type = DAE_OBJECT_PART;

		if(!core->region_id_ok(idx)){
			err=27;
			app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFYFAIL,0,0,0,(int*)&_trans_id,&zero,&zero,&err,0,0,0);
			printf(" [FAIL] <region id not ok (%i)>\n", idx);
			return;
		}
	}

	unsigned oid = 0; //FIXME this is sent in the msg

	// s4 : prop
	memcpy(&prop,&_data[4],4);
	if (debug) printf("songmsg_set_obj(): prop=%u (0x%04x)\n", prop, prop);

	if ((prop< 1) || ((prop> 6) && (prop<10)) || ((prop>17) && (prop<20)) ||
		((prop>51) && (prop<60)) ||((prop>65) && (prop<90)) || ((prop>96) && (prop<99)) || (prop>101))
		//	if ((prop<1) || (prop>45))
		{
		if (debug) printf("songmsg_set_obj(): bad property value: %u\n", prop);
		err=29;
		app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFYFAIL,0,0,0,(int*)&_trans_id,&zero,&zero,&err,0,0,0);
		printf(" [FAIL]\n");
		return;
	}
	// rest of data
	jack_nframes_t pos = 0;
	if (prop==1){
		//memcpy(&val_sp, &_data[8], 8);
		memcpy(&pos, &_data[8], 8); //FIXME we are copying 8 bytes into a 4 byte var.
		if (debug) printf("%s(): pos=%d\n", __func__, pos);
	}
	if (prop==2 || prop==PROP_SPLIT){
		memcpy(&val_u64, &_data[8], 8);
		if (debug) printf("%s(): val_u64=%Lu 0x%016Lx\n", __func__, val_u64, val_u64);
		pos = core->mu2samples(val_u64);
	}
	if ((prop>=3) && (prop<=5)){
		memcpy(&val_i,&_data[8], 4);
		val_si=(short int)val_i;
		if (debug) printf("val_i=%d  val_si=%d\n",val_i,val_si);
	}

	memcpy(&val_ui, &_data[8], 4);
	val_usi = val_ui&0xffff;
	val_uc  = val_ui&0xff;
	if (debug) printf("val_ui=%u  val_usi=%u  val_uc=%u\n",val_ui, val_usi, val_uc);

	double val_double;
	memcpy(&val_double, &_data[8], 8);
	if (debug>1) printf("val_double=%f\n", val_double);

	// check for LOCK on obj[] flags !! - except for case when being unlocked!!
	/*
	if ((_type!=LST) && ((dcore->obj[idx].emp.flags&0x10)>0)){
		if ((prop!=LOCKED) || (val_uc!=0))
			{
			err=26;
			app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFYFAIL,0,0,0,&_trans_id,&zero,&zero,&err,0,0,0);
			printf(" [FAIL]\n");
			return;
			}
	}
	*/

	switch (prop){
		case STIME:
			if (debug) printf("%s STIME%s\n", green, white);
			if (_type==DAE_OBJECT_PART){
			    if(debug>1) printf("type=PART\n");
			
				//find the region object
#ifdef NON_POD
				struct _region_shared* pod = (struct _region_shared*)shm_index->regions.slot[idx];
				uint64_t id = pod->id;
				AudioRegion* region = core->region_find(id);
#else
				//AudioRegion* region = core->playlist_region_find(idx);     //this function is unfortunately not yet working...

				//struct _region_shared* pod = (struct _region_shared*)shm_index->regions.slot[idx];
				/*struct block* current_block =*/ core->msg_idx_get_block(idx, &block_num, &m);
				printf("%s(): a=%i b=%i\n", __func__, block_num, m);


				/*
				//struct block* current_block = shm_index->regions.block[0];
				struct _region_shared* pod = (struct _region_shared*)shm_index->regions.block[0]->slot[idx];

				AudioRegion* region = (AudioRegion*)(pod->ardour_object);
				if (debug) printf("%s(): region_obj=%p name='%s'\n", __func__, region, region->name().c_str());
				*/
#endif
				/*
				if(region) {
					if (debug) printf("%s(): region.start=%u old_position=%u new_position=%u\n", __func__, region->start(), region->position(), pos);

                    //note that we need to subtract region->start due to the way ardour stores the position of the part.

					//session->add_undo (r.playlist()->get_memento());
					region->set_position(pos, NULL);
					//session->add_redo_no_execute (r.playlist()->get_memento());

					//AudioRegionEditor::bounds_changed (Change what_changed)
					err=0;
				}else err=1;//wrong error number.
				*/
				err = !DAE::core->region_set_start(block_num, m, pos);
			}
			else if (_type==DAE_OBJECT_EVENT){
			}
			else err=30;
			break;
		case LEN:
			if (debug) printf("%s LEN%s\n", green, white);
			if (_type==DAE_OBJECT_PART){
				/*struct block* current_block = */core->msg_idx_get_block(idx, &block_num, &m);
				err = !core->region_set_length(block_num, m, pos);
			}
			else err=31;
			break;
		case PROP_TRACK:
			if (debug) printf("%s TRACK%s\n", green, white);
			if (_type==DAE_OBJECT_PART){
				/*struct block* current_block = */core->msg_idx_get_block(idx, &block_num, &m);

				err = !core->region_set_track(block_num, m, val_ui);

#if 0   //moved
				AudioRegion* _region = (AudioRegion*)((struct _region_shared*)current_block->slot[idx])->ardour_object;
				boost::shared_ptr<ARDOUR::Region> region = (boost::shared_ptr<ARDOUR::Region>)_region;
				if(region){
					printf("  part=%i new_track=%i\n", idx, val_ui);

					//how to convert the "track" to playlist?
					//note: it looks like its possible that empty tracks/routes dont have any playlists!
					//ok, how we seem to be doing this, is to associate a route name eg "Audio 1" with a playlist name eg "Audio 1.1"
					//surely there are some nice functions for doing this? lets look...

					//remove the region from the old track:
					Playlist* old_playlist = region->playlist();
					if(old_playlist) old_playlist->remove_region(region);

					if(boost::shared_ptr<AudioTrack> track = core->get_track_boost(val_ui)){
						/*
						for(int i=0;i<10;i++){
							std::stringstream s;
							s << i;
							string playlist_name = i ? track->name() + "." + s.str() : track->name(); //try without appended number first. 
							Playlist* playlist = core->session->playlist_by_name(playlist_name);
							if(playlist){
								cout << "  setting playlist '" << playlist_name + "' = " << playlist << endl;
								playlist->add_region(region, region->position(), 1, false);
								err=0;
								return_idx = val_ui; //done!
								break;
							} else {
								cout <<  __func__ << "():  playlist not found (" << playlist_name << ")" << endl;
								err=1; //this is probably the wrong error
							}
						}
						*/
						Playlist* playlist = core->playlist_lookup_by_trackname(track->name());
						if(playlist){
							cout << "  setting playlist '" << playlist->name() + "' = " << playlist << endl;
							playlist->add_region(region, region->position(), 1, false);
							err=0;
							return_idx = val_ui; //done!
							break;
						} else {
							cout <<  __func__ << "():  playlist not found. track='" << track->name() << "'" << endl;
							err=1; //this is probably the wrong error
						}

					} else err=1;
				} else err=1; //this is probably the wrong error
#endif
			}
			break;
		case PROP_TRIM_LEFT:
			if (debug) printf("%s TRIM_LEFT%s\n", green, white);
			if (_type==DAE_OBJECT_PART){
				uint32_t left = core->mu2samples(val_ui);
				err = !core->region_trim_left(idx, left);

				/*  moved!
				struct block* current_block = core->msg_idx_get_block(idx, &block_num, &m);
				//AudioRegion* region = (AudioRegion*)((struct _region_shared*)shm_index->regions.slot[idx])->object;
				AudioRegion* region = (AudioRegion*)((struct _region_shared*)current_block->slot[idx])->ardour_object;
				if(region){
		
					//check existing region:
					//if(!region->verify_start_and_length(region->start(), region->length())) printf("  error! old region is too long!\n"); //private!!

					printf("  part=%i old_position=%u new_left=%u old_left=%i\n", idx, region->position(), left, region->position() + region->start());
					region->trim_front(left, NULL);
					//region->trim_start(left, NULL); //doesnt do much.
					//region->set_start(left - region->position(), NULL);  //verification errors - dont understand... but i guess this is not an appropriate function to use.
					printf("  position=%u new_left=%u\n", region->position(), region->position() + region->start());
				}
				*/
			}
			break;
		case PROP_SPLIT:
			if (debug) printf("%s SPLIT%s\n", green, white);
			if (_type==DAE_OBJECT_PART){
				uint32_t split_point = core->mu2samples(val_u64);
				err = !core->region_split(idx, split_point);

				//notify client:
				int notify_type = NOTIFYFAIL;
				if(!err) notify_type = NOTIFY;
				int obj_id = 0;
				int idx2   = 0; //FIXME we need to return the idx of the new region - no this isnt versatile enough.
				app_send_msg0(GARR, *frag, *off, HI, GENERAL, notify_type,0,0,0, (int*)&_trans_id, &obj_id, &idx2, &err,0,0,0);
				return;
			}
			break;
		case STRING:
			if (debug) printf("  %sSTRING%s", green, white);
			char str[64];
			memcpy(str, &_data[8], 63);
			str[63] = '\0';

			if (_type==DAE_OBJECT_PART){
				printf("  DAE_OBJECT_PART: string=%s\n", str);
				err = !core->region_set_name(idx, str);
			}
			else if (_type==DAE_OBJECT_SONG){
				//FIXME - temporarily for convenience, NEW_SONG arrives here!!
				printf("  DAE_OBJECT_SONG: string=%s\n", str);
				char snapshot[] = "1";
				if(core->load_session(str, snapshot, "/usr/local/share/ardourd/templates/2 Track.template")) err=1;
			}
			else err=85;
			break;
		case MUTE:
			if (debug) printf("  %sMUTE%s", green, white);
			if (_type==DAE_OBJECT_TRACK){
				printf("  track - setting:%i.\n", val_i);
				err = !core->route_set_mute(idx, val_i);
            }
			else err=85;
			break;
		case PROP_TEMPO:
			if (debug) printf("  %sTEMPO%s", green, white);
			core->set_tempo(val_double);
			break;

		case PROP_NEW_SONG:
			if (debug) printf("%s NEW_SONG%s\n", green, white);
			printf("  FIXME ???\n");

			//see STRING

			//note: core->load_session() includes unload_session().
#ifdef LOAD_EMPTY_SONG
			core->load_session("", "", "");
#else
			//-we have to load a template, as the gui doesnt like songs with no tracks.
			//char snapshot[256];
			//char mix_template[] = "2 Track.template";
			core->load_session("", "", "2 Track.template");
#endif

			break;
		case PROP_LOAD_SONG:
			if (debug) printf("%s LOAD_SONG%s\n", green, white);
			printf("  FIXME\n");
			char path[64];// = "FIXME";
			strcpy(path, " FIXME\n");
			core->load_session(path, "", "");
			break;
		case PROP_SAVE_SONG:
			if (debug) printf("%s SAVE_SONG%s\n", green, white);
			core->save();
			break;
		case PB_LEVEL:
			if (debug) printf("PB_LEV");
			// TRK or PRT
			switch(_type){
				case DAE_OBJECT_PART:
					printf("  DAE_OBJECT_PART level=%.2f\n", val_double);
					err = !core->region_set_pblevel(idx, val_double);
					break;
				default:
					printf("%s(): unhandled type for PB_LEVEL. type=%i\n", __func__, _type);
					err=34;
			}
			break;
		case ARM:
			if (debug) printf("ARM");
			switch(_type){
				case DAE_OBJECT_TRACK:
					printf("  DAE_OBJECT_TRACK arm_state=%i\n", val_i);
					err = !core->route_set_arm(idx, val_i);
					break;
				default:
					printf("%s(): ARM not valid for this object type. type=%i\n", __func__, _type);
					err=34;
			}
			break;
		/*
		case PB_DELAY_MU:
			if (debug) printf("PB_DELAY_MU");
// TRK or PRT
			if (_type==TRK) dcore->obj[idx].trk.pb_delay_mu=val_si;
			else if (_type==PRT) dcore->obj[idx].prt.pb_delay_mu=val_si;
			else err=32;
			break;
		case PB_DELAY_SAMPLE:
			if (debug) printf("PB_DELAY_SAMPLE");
// TRK or PRT
			if (_type==TRK) dcore->obj[idx].trk.pb_delay_sample=val_si;
			else if (_type==PRT) dcore->obj[idx].prt.pb_delay_sample=val_si;
			else err=33;
			break;
		case PB_PAN:
			if (debug) printf("PB_PAN");
			// TRK or PRT
			if (_type==TRK) dcore->obj[idx].trk.pb_pan=val_si;
			else if (_type==PRT) dcore->obj[idx].prt.pb_pan=val_si;
			else err=35;
			break;
		case REL_SP: // err=36
			if (_type!=LST) dcore->obj[idx].emp.flags=((dcore->obj[idx].emp.flags&0xfe) | (val_uc&0x01));
			else err=36;
			break;
		case SMUTE: // err=37
			if (_type!=LST) dcore->obj[idx].emp.flags=((dcore->obj[idx].emp.flags&0xfd) | ((val_uc&0x01)<<1));
			else err=37;
			break;
		case SSOLO: // err=38
			if (_type!=LST) dcore->obj[idx].emp.flags=((dcore->obj[idx].emp.flags&0xfb) | ((val_uc&0x01)<<2));
			else err=38;
			break;
		case RECORD: // err=39
			if (_type!=LST) dcore->obj[idx].emp.flags=((dcore->obj[idx].emp.flags&0xf7) | ((val_uc&0x01)<<3));
			else err=39;
			break;
		case LOCKED: // err=40
			if (_type!=LST) dcore->obj[idx].emp.flags=((dcore->obj[idx].emp.flags&0xef) | ((val_uc&0x01)<<4));
			else err=40;
			break;
		case PARENT: // err=41
			if (_type!=STR) 
				{
				if (val_ui>0) add_child_to_parent(idx,val_ui);
				else remove_child_from_parent(idx);
//				dcore->obj[idx].trk.parent=val_ui;
				}
			else err=41;
			break;
		case MOTHER: // err=42
			if (_type!=STR)
				{
				if (val_ui>0) add_child_to_mother(idx,val_ui);
				else dcore->obj[idx].trk.mother=val_ui;
				}
			else err=42;
			break;
		case FOLDER: // err=43
			if (debug) printf("FOLDER");
// PRT
//			if (_type==PRT)
			if (_type!=LST) dcore->obj[idx].prt.flags=((dcore->obj[idx].prt.flags&0xdf) | ((val_uc&0x01)<<5));
			else err=43;
			break;
		case COLOUR:
			if (debug) printf("COLOUR");
			// TRK or CHA or PRT or EVT
			if (_type==TRK) dcore->obj[idx].trk.colour=val_ui;
			else if (_type==CHA) dcore->obj[idx].cha.colour=val_ui;
			else if (_type==PRT) dcore->obj[idx].prt.colour=val_ui;
			else if (_type==EVT) dcore->obj[idx].evt.colour=val_ui;
			else err=44;
			break;
		case HEIGHT:
			if (debug) printf("HEIGHT");
			// TRK
			if (_type==TRK) dcore->obj[idx].trk.height=val_ui;
			else err=45;
			break;
		case NUM:
			if (debug) printf("NUM");
			// TRK or CHA
			if (_type==TRK) dcore->obj[idx].trk.trk_num=val_ui;
			else if (_type==CHA) dcore->obj[idx].cha.chan_num=val_ui;
			else err=46;
			break;
		case EVENT_TYPE:
			if (debug) printf("EVENT_TYPE");
			// EVT
			if (_type==EVT)
				{
				     if (_sub==MIDI) dcore->obj[idx].evt.ev.mer.event_type=val_uc;
				else if (_sub==SIMP) dcore->obj[idx].evt.ev.ser.event_type=val_uc;
				else if (_sub==PLUG) dcore->obj[idx].evt.ev.per.event_type=val_uc;
				else if (_sub==MIXE) dcore->obj[idx].evt.ev.xer.event_type=val_uc;
				else if (_sub==OPER) dcore->obj[idx].evt.ev.oer.event_type=val_uc;
				else err=47;
				}
			else err=47;
			break;
		case SAMPLE_START:
			if (debug) printf("SAMPLE_START");
// EVT : AUDI
// EVT : SIMP
			if ((_type==EVT) && (_sub==AUDI)) dcore->obj[idx].evt.ev.aer.sample_start=val_ui;
			else if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.sample_start=val_ui;
			else err=48;
			break;
		case SAMPLE_END:
			if (debug) printf("SAMPLE_END");
// EVT : AUDI
// EVT : SIMP
			if ((_type==EVT) && (_sub==AUDI)) dcore->obj[idx].evt.ev.aer.sample_end=val_ui;
			else if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.sample_end=val_ui;
			else err=49;
			break;
		case PTR_IDX:
			if (debug) printf("PTR_IDX (fil_idx or label_idx)");
// label_idx: TRK, CHA, PRT, FILE
// fil_idx:
// EVT : AUDI
// EVT : SIMP
// EVT : VIDE
			     if (_type==TRK) dcore->obj[idx].trk.label_idx=val_ui;
			else if (_type==CHA) dcore->obj[idx].cha.label_idx=val_ui;
			else if (_type==PRT) dcore->obj[idx].prt.label_idx=val_ui;
			else if (_type==FIL) dcore->obj[idx].fil.label_idx=val_ui;
			else if ((_type==EVT) && (_sub==AUDI)) dcore->obj[idx].evt.ev.aer.fil_idx=val_usi;
			else if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.fil_idx=val_usi;
			else err=50;

			break;
		case MTYPE:
			if (debug) printf("MTYPE");
// EVT : MIDI
			if ((_type==EVT) && (_sub==MIDI)) dcore->obj[idx].evt.ev.mer.type=val_uc;
			else err=51;
			break;
		case MCHAN:
			if (debug) printf("MCHAN");
// EVT : MIDI
			if ((_type==EVT) && (_sub==MIDI)) dcore->obj[idx].evt.ev.mer.chan=val_uc;
			else err=52;
			break;
		case MNOTE:
			if (debug) printf("MNOTE");
// EVT : MIDI
// EVT : SIMP
			if ((_type==EVT) && (_sub==MIDI)) dcore->obj[idx].evt.ev.mer.note=val_uc;
			else if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.note=val_uc;
			else err=53;
			break;
		case VEL:
			if (debug) printf("VEL");
// EVT : MIDI
// EVT : SIMP
			if ((_type==EVT) && (_sub==MIDI)) dcore->obj[idx].evt.ev.mer.vel=val_uc;
			else if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.vel=val_uc;
			else err=54;
			break;
		case OFF_VEL:
			if (debug) printf("OFF_VEL");
// EVT : MIDI
			if ((_type==EVT) && (_sub==MIDI)) dcore->obj[idx].evt.ev.mer.off_vel=val_uc;
			else err=55;
			break;
		case DUR:
			if (debug) printf("DUR");
// EVT : MIDI
			if ((_type==EVT) && (_sub==MIDI)) dcore->obj[idx].evt.ev.mer.dur=val_uc;
			else err=56;
			break;
		case ENVELOPE:
			if (debug) printf("ENVELOPE");
// EVT : SIMP
			if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.envelope=val_uc;
			else err=57;
			break;
		case LOOP:
			if (debug) printf("LOOP");
// EVT : SIMP
			if ((_type==EVT) && (_sub==SIMP)) dcore->obj[idx].evt.ev.ser.loop=val_uc;
			else err=58;
			break;
		case ACHAN:
			if (debug) printf("%sACHAN%s\n", green, white);
// EVT : MIXE
			if ((_type==EVT) && (_sub==MIXE)) dcore->obj[idx].evt.ev.xer.channel=val_ui;
			else err=59;
			break;
		case VALUE:
			if (debug) printf("VALUE");
// EVT : MIXE
			if ((_type==EVT) && (_sub==MIXE)) dcore->obj[idx].evt.ev.xer.value=val_ui;
			else err=60;
			break;
		case CTL_IDX:
			if (debug) printf("CTL_IDX");
// EVT : MIXE
			if ((_type==EVT) && (_sub==MIXE)) dcore->obj[idx].evt.ev.xer.ctl_idx=val_usi;
			else err=61;
			break;
		case OP_TYPE:
			if (debug) printf("OP_TYPE");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.op_type=val_uc;
			else err=62;
			break;
		case FUNCTION:
			if (debug) printf("FUNCTION");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.function=val_uc;
			else err=63;
			break;
		case TRIGGER:
			if (debug) printf("TRIGGER");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.trigger=val_uc;
			else err=64;
			break;
		case ARGS:
			if (debug) printf("ARGS");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.args=val_uc;
			else err=65;
			break;
		case TARGET:
			if (debug) printf("TARGET");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.target=val_uc;
			else err=66;
			break;
		case OUTPUT:
			if (debug) printf("OUTPUT");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.output=val_uc;
			else err=67;
			break;
		case CHAIN:
			if (debug) printf("CHAIN");
// EVT : OPER
			if ((_type==EVT) && (_sub==OPER)) dcore->obj[idx].evt.ev.oer.chain=val_uc;
			else err=68;
			break;
		case FRAME_NUM:
			if (debug) printf("FRAME_NUM");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.frame_num=val_ui;
			else err=69;
			break;
		case VIDEO_FMT:
			if (debug) printf("VIDEO_FMT");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.video_fmt=val_uc;
			else err=70;
			break;
		case COMP_FMT:
			if (debug) printf("COMP_FMT");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.comp_fmt=val_uc;
			else err=71;
			break;
		case VHEIGHT:
			if (debug) printf("VHEIGHT");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.height=val_usi;
			else err=72;
			break;
		case VWIDTH:
			if (debug) printf("VWIDTH");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.width=val_usi;
			else err=73;
			break;
		case DWIDTH:
			if (debug) printf("DWIDTH");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.dwidth=val_usi;
			else err=74;
			break;
		case DHEIGHT:
			if (debug) printf("DHEIGHT");
// EVT : VIDE
			if ((_type==EVT) && (_sub==VIDE)) dcore->obj[idx].evt.ev.ver.dheight=val_usi;
			else err=75;
			break;

		case INTERLEAVE:
			if (debug) printf("INTERLEAVE");
// FIL
			if (_type==FIL) dcore->obj[idx].fil.interleave=val_uc;
			else err=76;
			break;
		case BITS:
			if (debug) printf("BITS");
// FIL
			if (_type==FIL) dcore->obj[idx].fil.bits=val_uc;
			else err=77;
			break;

// ALL part of fflags .....
		case INT_FLOAT:
			if (debug) printf("INT_FLOAT");
// FIL
			if (_type==FIL) dcore->obj[idx].fil.fflags=((dcore->obj[idx].fil.fflags&0xfe) | (val_uc&0x01));
			else err=78;
			break;
		case CPU_PCM:
			if (debug) printf("CPU_PCM");
// FIL
			if (_type==FIL) dcore->obj[idx].fil.fflags=((dcore->obj[idx].fil.fflags&0xfd) | ((val_uc&0x01)<<1));
			else err=79;
			break;
		case LE_BE:
			if (debug) printf("LE_BE");
// FIL
			if (_type==FIL) dcore->obj[idx].fil.fflags=((dcore->obj[idx].fil.fflags&0xfb) | ((val_uc&0x01)<<2));
			else err=80;
			break;
// *** END of fflags ***

		case SAMPLE_RATE:
			if (debug) printf("SAMPLE_RATE");
// FIL
			if (_type==FIL) dcore->obj[idx].fil.sample_rate=val_ui;
			else err=81;
			break;

		case REPEAT_COUNT:
			if (debug) printf("REPEAT_COUNT");
// PRT
			if (_type==PRT) dcore->obj[idx].prt.repeat_count=val_usi;
			else err=82;
			break;

		case REPEAT_LEN:
			if (debug) printf("REPEAT_LEN");
// PRT
			if (_type==PRT) dcore->obj[idx].prt.repeat_len=val_ull;
			else err=83;
			break;

		case TAG:
			if (debug) printf("TAG");
// RAW
			if (_type==RAW) dcore->obj[idx].raw.tag=val_uc;
			else err=84;
			break;

		case CHAN:
			{
			if (debug) printf("  chan!! val_i=%i val_ui=%u\n\n\n", val_i, val_ui);
			//would this work?: add_child_to_mother(idx,val_ui);

			//if the track has no existing children we use its Child prop. Else we the the child's Next prop. 
			//if(dcore->obj[idx].emp.type==TRK) dcore->obj[idx].trk.child = val_ui;
			//else 
			dcore->obj[idx].trk.next  = val_ui;
			break;
			}
		*/
		default:
			err=29;
	}
	printf(" ");
	printf("%s(): sending notification...\n ", __func__);
	if (err==0){
		//int int_idx = GPOINTER_TO_INT(idx);
		app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFY,0,0,0,(int*)&_trans_id, (int*)&oid, &return_idx, &err,0,0,0);
		//if (debug) printf(" [OK]\n");
	}
	else{
		app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFYFAIL,0,0,0,(int*)&_trans_id,&zero,&zero,&err,0,0,0);
		//printf(" [%sFAIL%s]\n", red, white);
		printf(" %s\n", fail);
	}
}


struct _msg_add_part{
    char     msg_type;
    char     sub_type;
    char     colour;
    char     from_region;
    uint32_t trk_pod_idx;    //4
    uint32_t mother;         //8
    uint64_t len;            //12
    struct _song_pos stime;  //20
    uint16_t repeat_count;   //28  -this is taking 4 bytes! items below here are off!
    uint64_t repeat_len;     //30
    uint32_t source;         //38  -can either be a file or a region.
    short    pb_level;       //42
    short    pb_pan;         //44
    char     name[50];       //46
};
int
dae_msg_add_part(unsigned trans_id, int* frag, short int* off, char *data)
{
	int err = 0;
	struct _msg_add_part* msg_string = (struct _msg_add_part*)data;

	bool from_region = msg_string->from_region; //the new part is either based on a file or an existing region.
	printf("%s(): from_region=%01i\n", __func__, from_region);

	string name(&data[46], 0, 64);

	//find which region this new part is based on.
	uint32_t src_region_idx;
	memcpy(&src_region_idx, &data[38], 4);

	uint32_t trk_pod_idx = msg_string->trk_pod_idx;

	jack_nframes_t start = core->pos2samples(&msg_string->stime);

	uint64_t len_mu = 0;
	memcpy(&len_mu, &data[12], 8);
	jack_nframes_t length = core->mu2samples(len_mu);

	printf("%s(): beat=%i start=%u.\n", __func__, msg_string->stime.beat, start);

    uint32_t new_obj_id;
	uint32_t new_obj_idx;
	int r = core->add_part (name, from_region, src_region_idx, trk_pod_idx, start, length, &new_obj_id, &new_obj_idx);
    if(r) app_send_msg0(GARR, *frag, *off, HI, GENERAL, NOTIFY,0,0,0, (int*)&trans_id, (int*)&new_obj_id, (int*)&new_obj_idx, &err,0,0,0);
	return r;
}

int
dae_msg_del_part(unsigned trans_id, int pod_idx)
{
	printf("%s()...\n", __func__);
	int err = core->del_part(pod_idx);
	if(!err){
		int obj_id = 0;
		app_send_msg0(GARR, 0,0, HI, GENERAL, NOTIFY,0,0,0, (int*)&trans_id, &obj_id, &pod_idx, &err,0,0,0);
	} else {
		app_send_msg0(GARR, 0,0, HI, GENERAL, NOTIFYFAIL,0,0,0, (int*)&trans_id, &zero, &zero, &err,0,0,0);
	}
	return err;
}


int
dae_msg_add_track(unsigned int _trans_id, int *frag, short int *off, char *_data)
{
	uint32_t new_obj_id, new_obj_idx;

	string name(&_data[22], 0, 64);

	int err = core->add_track(name, &new_obj_idx, &new_obj_id);
	int obj_id  = (int)new_obj_id;
	int obj_idx = (int)new_obj_idx;
	if(!err) app_send_msg0(GARR, *frag, *off, HI, GENERAL, NOTIFY,0,0,0, (int*)&_trans_id, &obj_id, &obj_idx, &err,0,0,0);
	else     app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFYFAIL,0,0,0, (int*)&_trans_id, &zero, &zero, &err,0,0,0);

	return err;
}


int
dae_msg_del_track(unsigned trans_id, int pod_idx)
{
	printf("%s()...\n", __func__);

	int obj_id = 0;
	int err = core->del_track(pod_idx);

	app_send_msg0(GARR, 0,0, HI, GENERAL, NOTIFY,0,0,0, (int*)&trans_id, &obj_id, &pod_idx, &err,0,0,0);

	return err;
}


int
dae_msg_add_file(unsigned _trans_id, int *frag, short int *off, char *_data)
{
	//FIXME our filepath string is not long enough!

	printf("%s()...\n", __func__);
	if(debug) printf("%s(%d,%d,%d,...)\n", __func__, _trans_id, *frag, *off);

	int idx=0;
	unsigned obj_id=0;

	char path[64];
	memset(path, 0, 52);
	snprintf(path, 50, "%s", &_data[18]);
	if(debug) printf("name = >%s<\n", path);
	string path_str(path);

	int err = core->add_file(path_str);

	if(1){
		printf("%s(): ok!\n", __func__);
	}else{
		printf("** ERROR ** obj_id=\n");
		err=8;
		app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFYFAIL,0,0,0, (int*)&_trans_id, &zero, &zero, &err,0,0,0);
		return 1;
	}

	app_send_msg0(GARR,*frag,*off,HI,GENERAL,NOTIFY,0,0,0, (int*)&_trans_id, (int*)&obj_id, &idx, &err, 0,0,0);

	return err;
}


int
dae_msg_del_file(unsigned trans_id, int pod_idx)
{
	printf("%s()...\n", __func__);

	int obj_id = 0;
	int err = core->del_file(pod_idx);
	app_send_msg0(GARR, 0,0, HI, GENERAL, NOTIFY,0,0,0, (int*)&trans_id, &obj_id, &pod_idx, &err,0,0,0);

	return true;
}


void
dae_msg_set_output(unsigned _trans_id, int* frag, short int* off, unsigned char* msg)
{
	printf("%s()...\n", __func__);
	unsigned short int tsint1=0, tsint2=0;

	memcpy(&tsint1, &msg[0], 2);     //bus
	//if (debug) printf("bus=%u ",tsint1);
	unsigned ival1 = (unsigned int)tsint1;
	//struct _connection_shared* cshared = (struct _connection_shared*)shm_index->route_index[ival1];

	memcpy(&tsint2, &msg[2], 2);     //track
	unsigned tnum = (unsigned)tsint2;
	printf("%s(): tnum=%u\n", __func__, tnum);

	//send notification:
	int ival3=0;
	app_send_msg(GMIXER,0,0,HI,EVENT,NOTIFY,BS,0,0,(int*)&ival1,(int*)&tnum, &ival3,0,0,0,0,0,0);
}


} //end namespace DAE


void
msg_printf(unsigned char* msg)
{
	int c;
	for(c=0;c<32;c++){
		printf("%s%02x", c%4 ? " " : " - ", msg[c]);
	}
	printf("\n");
}


